from pathlib import Path
from piptools.cache import DependencyCache
from piptools.repositories import PyPIRepository
from piptools.resolver import Resolver
from piptools.scripts.compile import get_pip_command

from cachy import CacheManager

import poetry.packages

from poetry.locations import CACHE_DIR
from poetry.packages import dependency_from_pep_508
from poetry.semver.constraints import Constraint
from poetry.semver.constraints.base_constraint import BaseConstraint
from poetry.semver.version_parser import VersionParser
from poetry.version.markers import InvalidMarker

from .pypi_repository import PyPiRepository


class LegacyRepository(PyPiRepository):

    def __init__(self, name, url):
        if name == 'pypi':
            raise ValueError('The name [pypi] is reserved for repositories')

        self._name = name
        self._url = url
        command = get_pip_command()
        opts, _ = command.parse_args([])
        self._session = command._build_session(opts)
        self._repository = PyPIRepository(opts, self._session)
        self._cache_dir = Path(CACHE_DIR) / 'cache' / 'repositories' / name

        self._cache = CacheManager({
            'default': 'releases',
            'serializer': 'json',
            'stores': {
                'releases': {
                    'driver': 'file',
                    'path': Path(CACHE_DIR) / 'cache' / 'repositories' / name
                },
                'packages': {
                    'driver': 'dict'
                },
                'matches': {
                    'driver': 'dict'
                }
            }
        })

    def find_packages(self, name, constraint=None, extras=None):
        packages = []

        if constraint is not None and not isinstance(constraint,
                                                     BaseConstraint):
            version_parser = VersionParser()
            constraint = version_parser.parse_constraints(constraint)

        key = name
        if constraint:
            key = f'{key}:{str(constraint)}'

        if self._cache.store('matches').has(key):
            versions = self._cache.store('matches').get(key)
        else:
            candidates = [str(c.version) for c in self._repository.find_all_candidates(name)]

            versions = []
            for version in candidates:
                if version in versions:
                    continue

                if (
                    not constraint
                    or (constraint and constraint.matches(Constraint('=', version)))
                ):
                    versions.append(version)

            self._cache.store('matches').put(key, versions, 5)

        for version in versions:
            packages.append(self.package(name, version, extras=extras))

        return packages

    def package(self, name, version, extras=None) -> 'poetry.packages.Package':
        """
        Retrieve the release information.

        This is a heavy task which takes time.
        We have to download a package to get the dependencies.
        We also need to download every file matching this release
        to get the various hashes.
        
        Note that, this will be cached so the subsequent operations
        should be much faster.
        """
        try:
            index = self._packages.index(
                poetry.packages.Package(name, version, version)
            )

            return self._packages[index]
        except ValueError:
            if extras is None:
                extras = []

            release_info = self.get_release_info(name, version)
            package = poetry.packages.Package(name, version, version)
            for req in release_info['requires_dist']:
                try:
                    dependency = dependency_from_pep_508(req)
                except InvalidMarker:
                    # Invalid marker
                    # We strip the markers hoping for the best
                    req = req.split(';')[0]

                    dependency = dependency_from_pep_508(req)

                if dependency.extras:
                    for extra in dependency.extras:
                        if extra not in package.extras:
                            package.extras[extra] = []

                        package.extras[extra].append(dependency)

                if not dependency.is_optional():
                    package.requires.append(dependency)

            # Adding description
            package.description = release_info.get('summary', '')

            # Adding hashes information
            package.hashes = release_info['digests']

            # Activate extra dependencies
            for extra in extras:
                if extra in package.extras:
                    for dep in package.extras[extra]:
                        dep.activate()

                    package.requires += package.extras[extra]

            self._packages.append(package)

            return package

    def get_release_info(self, name: str, version: str) -> dict:
        """
        Return the release information given a package name and a version.

        The information is returned from the cache if it exists
        or retrieved from the remote server.
        """
        return self._cache.store('releases').remember_forever(
            f'{name}:{version}',
            lambda: self._get_release_info(name, version)
        )

    def _get_release_info(self, name: str, version: str) -> dict:
        from pip.req import InstallRequirement
        from pip.exceptions import InstallationError

        ireq = InstallRequirement.from_line(f'{name}=={version}')
        resolver = Resolver(
            [ireq], self._repository,
            cache=DependencyCache(self._cache_dir.as_posix())
        )
        try:
            requirements = list(resolver._iter_dependencies(ireq))
        except InstallationError as e:
            # setup.py egg-info error most likely
            # So we assume no dependencies
            requirements = []

        requires = []
        for dep in requirements:
            constraint = str(dep.req.specifier)
            require = f'{dep.name}'
            if constraint:
                require += f' ({constraint})'

            requires.append(require)

        hashes = resolver.resolve_hashes([ireq])[ireq]
        hashes = [h.split(':')[1] for h in hashes]

        data = {
            'name': name,
            'version': version,
            'summary': '',
            'requires_dist': requires,
            'digests': hashes
        }

        resolver.repository.freshen_build_caches()

        return data



import re

from poetry.semver.constraints import Constraint
from poetry.semver.constraints.base_constraint import BaseConstraint
from poetry.semver.version_parser import VersionParser

from poetry.version import parse as parse_version

from .base_repository import BaseRepository


class Repository(BaseRepository):

    def __init__(self, packages=None):
        super(Repository, self).__init__()

        if packages is None:
            packages = []

        for package in packages:
            self.add_package(package)

    def package(self, name, version):
        name = name.lower()
        version = str(parse_version(version))

        for package in self.packages:
            if name == package.name and package.version == version:
                return package

    def find_packages(self, name, constraint=None, extras=None):
        name = name.lower()
        packages = []
        if extras is None:
            extras = []

        if not isinstance(constraint, BaseConstraint):
            parser = VersionParser()
            constraint = parser.parse_constraints(constraint)

        for package in self.packages:
            if name == package.name:
                pkg_constraint = Constraint('==', package.version)

                if constraint is None or constraint.matches(pkg_constraint):
                    for extra in extras:
                        if extra in package.extras:
                            for dep in package.extras[extra]:
                                dep.activate()

                            package.requires += package.extras[extra]

                    packages.append(package)

        return packages

    def search(self, query, mode=0):
        regex = '(?i)(?:{})'.format('|'.join(re.split('\s+', query)))

        matches = {}
        for package in self.packages:
            name = package.name

            if name in matches:
                continue

            if (
                re.match(regex, name) is not None
                or (
                    mode == self.SEARCH_FULLTEXT
                    and isinstance(package, CompletePackage)
                    and re.match(regex, '')
                )
            ):
                matches[name] = {
                    'name': package.pretty_name,
                    'description': (package.description
                                    if isinstance(package, CompletePackage)
                                    else '')
                }

        return list(matches.values())

    def has_package(self, package):
        package_id = package.unique_name

        for repo_package in self.packages:
            if package_id == repo_package.unique_name:
                return True

        return False

    def add_package(self, package):
        self._packages.append(package)

    def remove_package(self, package):
        package_id = package.unique_name

        index = None
        for i, repo_package in enumerate(self.packages):
            if package_id == repo_package.unique_name:
                index = i
                break

        if index is not None:
            del self._packages[index]

    def __len__(self):
        return len(self._packages)



import toml

from poetry.utils.helpers import module_name
from poetry.vcs.git import Git


TESTS_DEFAULT = """from {package_name} import __version__


def test_version():
    assert '{version}' == __version__
"""


class Layout(object):

    def __init__(self, project, version='0.1.0', readme_format='md', author=None):
        self._project = project
        self._package_name = module_name(project)
        self._version = version
        self._readme_format = readme_format
        self._dependencies = {}
        self._dev_dependencies = {}
        self._include = []

        self._git = Git()
        git_config = self._git.config
        if not author:
            if (
                git_config.get('user.name')
                and git_config.get('user.email')
            ):
                author = '{} <{}>'.format(
                    git_config['user.name'],
                    git_config['user.email']
                )
            else:
                author = 'Your Name <you@example.com>'

        self._author = author

    def create(self, path, with_tests=True):
        self._dependencies = {}
        self._dev_dependencies = {}
        self._include = []

        path.mkdir(parents=True, exist_ok=True)

        self._create_default(path)
        self._create_readme(path)

        if with_tests:
            self._create_tests(path)

        self._write_poetry(path)

    def _create_default(self, path, src=True):
        raise NotImplementedError()

    def _create_readme(self, path):
        if self._readme_format == 'rst':
            readme_file = path / 'README.rst'
        else:
            readme_file = path / 'README.md'

        readme_file.touch()

    def _create_tests(self, path):
        self._dev_dependencies['pytest'] = '^3.0'

        tests = path / 'tests'
        tests_init = tests / '__init__.py'
        tests_default = tests / 'test_{}.py'.format(self._package_name)

        tests.mkdir()
        tests_init.touch(exist_ok=False)

        with tests_default.open('w') as f:
            f.write(
                TESTS_DEFAULT.format(
                    package_name=self._package_name,
                    version=self._version
                )
            )

    def _write_poetry(self, path):
        output = {
            'tool': {
                'poetry': {
                    'name': self._project,
                    'version': self._version,
                    'authors': [self._author],
                }
            }
        }

        content = toml.dumps(output, preserve=True)

        output = {
            'tool': {
                'poetry': {
                    'dependencies': {},
                    'dev-dependencies': {
                        'pytest': '^3.4'
                    }
                }
            }
        }

        content += '\n' + toml.dumps(output, preserve=True)

        poetry = path / 'pyproject.toml'

        with poetry.open('w') as f:
            f.write(content)



from .base_installer import BaseInstaller


class NoopInstaller(BaseInstaller):

    def __init__(self):
        self._installs = []
        self._updates = []
        self._removals = []

    @property
    def installs(self):
        return self._installs

    @property
    def updates(self):
        return self._updates

    @property
    def removals(self):
        return self._removals

    def install(self, package) -> None:
        self._installs.append(package)

    def update(self, source, target) -> None:
        self._updates.append((source, target))

    def remove(self, package) -> None:
        self._removals.append(package)



from poetry.utils.helpers import canonicalize_name
from poetry.version.helpers import format_python_constraint


class Metadata:

    metadata_version = '2.1'
    # version 1.0
    name = None
    version = None
    platforms = ()
    supported_platforms = ()
    summary = None
    description = None
    keywords = None
    home_page = None
    download_url = None
    author = None
    author_email = None
    license = None
    # version 1.1
    classifiers = ()
    requires = ()
    provides = ()
    obsoletes = ()
    # version 1.2
    maintainer = None
    maintainer_email = None
    requires_python = None
    requires_external = ()
    requires_dist = ()
    provides_dist = ()
    obsoletes_dist = ()
    project_urls = ()

    # Version 2.1
    description_content_type = None
    provides_extra = []

    @classmethod
    def from_package(cls, package) -> 'Metadata':
        meta = cls()

        meta.name = canonicalize_name(package.name)
        meta.version = package.version
        meta.summary = package.description
        if package.readme:
            with package.readme.open() as f:
                meta.description = f.read()

        meta.keywords = ','.join(package.keywords)
        meta.home_page = package.homepage or package.repository_url
        meta.author = package.author_name
        meta.author_email = package.author_email
        meta.license = package.license
        meta.classifiers = package.classifiers

        # Version 1.2
        meta.maintainer = meta.author
        meta.maintainer_email = meta.author_email
        meta.requires_python = package.python_constraint
        meta.requires_dist = [d.to_pep_508() for d in package.requires]

        # Requires python
        meta.requires_python = format_python_constraint(package.python_constraint)

        # Version 2.1
        if package.readme:
            if package.readme.suffix == '.rst':
                meta.description_content_type = 'text/x-rst'
            elif package.readme.suffix in ['.md', '.markdown']:
                meta.description_content_type = 'text/markdown'
            else:
                meta.description_content_type = 'text/plain'

        # TODO: Provides extra
        return meta



import os
import platform
import subprocess
import sys
import sysconfig
import warnings

from contextlib import contextmanager
from pathlib import Path
from subprocess import CalledProcessError

from venv import EnvBuilder

from poetry.config import Config
from poetry.locations import CACHE_DIR


class VenvError(Exception):

    pass


class VenvCommandError(VenvError):

    def __init__(self, e: CalledProcessError):
        message = f'Command {e.cmd} errored with the following output: \n' \
                  f'{e.output.decode()}'

        super().__init__(message)


class Venv:

    def __init__(self, venv=None):
        self._venv = venv
        if self._venv:
            self._venv = Path(self._venv)


        self._windows = sys.platform == 'win32'

        self._bin_dir = None
        if venv:
            bin_dir = 'bin' if not self._windows else 'Scripts'
            self._bin_dir = self._venv / bin_dir

        self._version_info = None
        self._python_implementation = None

    @classmethod
    def create(cls, io, name=None) -> 'Venv':
        if 'VIRTUAL_ENV' not in os.environ:
            # Not in a virtualenv
            # Checking if we need to create one
            config = Config.create('config.toml')

            create_venv = config.setting('settings.virtualenvs.create')

            venv_path = config.setting('settings.virtualenvs.path')
            if venv_path is None:
                venv_path = Path(CACHE_DIR) / 'virtualenvs'
            else:
                venv_path = Path(venv_path)

            if not name:
                name = Path.cwd().name

            name = f'{name}-py{".".join([str(v) for v in sys.version_info[:2]])}'

            venv = venv_path / name
            if not venv.exists():
                if create_venv is False:
                    io.writeln(
                        '<fg=black;bg=yellow>'
                        'Skipping virtualenv creation, '
                        'as specified in config file.'
                        '</>'
                    )

                    return cls()

                io.writeln(
                    f'Creating virtualenv <info>{name}</> in {str(venv_path)}'
                )
                builder = EnvBuilder(with_pip=True)
                builder.create(str(venv))
            else:
                if io.is_very_verbose():
                    io.writeln(f'Virtualenv <info>{name}</> already exists.')

            os.environ['VIRTUAL_ENV'] = str(venv)

        # venv detection:
        # stdlib venv may symlink sys.executable, so we can't use realpath.
        # but others can symlink *to* the venv Python,
        # so we can't just use sys.executable.
        # So we just check every item in the symlink tree (generally <= 3)
        p = os.path.normcase(sys.executable)
        paths = [p]
        while os.path.islink(p):
            p = os.path.normcase(
                os.path.join(os.path.dirname(p), os.readlink(p)))
            paths.append(p)

        p_venv = os.path.normcase(os.environ['VIRTUAL_ENV'])
        if any(p.startswith(p_venv) for p in paths):
            # Running properly in the virtualenv, don't need to do anything
            return cls()

        venv = os.environ['VIRTUAL_ENV']

        return cls(venv)

    @property
    def venv(self):
        return self._venv

    @property
    def python(self) -> str:
        """
        Path to current python executable
        """
        return self._bin('python')

    @property
    def pip(self) -> str:
        """
        Path to current pip executable
        """
        return self._bin('pip')

    @property
    def version_info(self) -> tuple:
        if self._version_info is not None:
            return self._version_info

        if not self.is_venv():
            self._version_info = sys.version_info
        else:
            output = self.run('python', '--version', shell=True)

            version = output.split(' ')
            self._version_info = tuple([
                int(s) for s in version[1].strip().split('.')
            ])

        return self._version_info

    @property
    def python_implementation(self):
        if self._python_implementation is not None:
            return self._python_implementation

        if not self.is_venv():
            impl = platform.python_implementation()
        else:
            impl = self.run(
                'python', '-c',
                '"import platform; print(platform.python_implementation())"',
                shell=True
            ).strip()

        self._python_implementation = impl

        return self._python_implementation

    def config_var(self, var):
        if not self.is_venv():
            try:
                return sysconfig.get_config_var(var)
            except IOError as e:
                warnings.warn("{0}".format(e), RuntimeWarning)
                return None

        try:
            value = self.run(
                'python', '-c',
                f'"import sysconfig; print(sysconfig.get_config_var(\'{var}\'))"',
                shell=True
            ).strip()
        except VenvCommandError as e:
            warnings.warn("{0}".format(e), RuntimeWarning)
            return None

        if value == 'None':
            value = None
        elif value == '1':
            value = 1
        elif value == '0':
            value = 0

        return value

    def run(self, bin: str, *args, **kwargs) -> str:
        """
        Run a command inside the virtual env.
        """
        cmd = [bin] + list(args)
        shell = kwargs.get('shell', False)

        if shell:
            cmd = ' '.join(cmd)

        try:
            if not self.is_venv():
                output = subprocess.check_output(
                    cmd, stderr=subprocess.STDOUT,
                    **kwargs
                )
            else:
                if self._windows:
                    kwargs['shell'] = True

                with self.temp_environ():
                    os.environ['PATH'] = self._path()
                    os.environ['VIRTUAL_ENV'] = str(self._venv)

                    self.unset_env('PYTHONHOME')
                    self.unset_env('__PYVENV_LAUNCHER__')

                    output = subprocess.check_output(
                        cmd, stderr=subprocess.STDOUT,
                        **kwargs
                    )
        except CalledProcessError as e:
            raise VenvCommandError(e)

        return output.decode()

    def exec(self, bin, *args, **kwargs):
        if not self.is_venv():
            return subprocess.run([bin] + list(args)).returncode
        else:
            with self.temp_environ():
                os.environ['PATH'] = self._path()
                os.environ['VIRTUAL_ENV'] = str(self._venv)

                self.unset_env('PYTHONHOME')
                self.unset_env('__PYVENV_LAUNCHER__')

                completed = subprocess.run([bin] + list(args), **kwargs)

                return completed.returncode

    @contextmanager
    def temp_environ(self):
        environ = dict(os.environ)
        try:
            yield
        finally:
            os.environ.clear()
            os.environ.update(environ)

    def _path(self):
        return os.pathsep.join([
            str(self._bin_dir),
            os.environ['PATH'],
        ])

    def unset_env(self, key):
        if key in os.environ:
            del os.environ[key]

    def get_shell(self):
        shell = Path(os.environ.get('SHELL', '')).stem
        if shell in ('bash', 'zsh', 'fish'):
            return shell

    def _bin(self, bin) -> str:
        """
        Return path to the given executable.
        """
        if not self.is_venv():
            return bin

        return str(self._bin_dir / bin) + ('.exe' if self._windows else '')

    def is_venv(self) -> bool:
        return self._venv is not None


class NullVenv(Venv):

    def __init__(self, execute=False):
        super().__init__()

        self.executed = []
        self._execute = execute

    def run(self, bin: str, *args):
        self.executed.append([bin] + list(args))

        if self._execute:
            return super().run(bin, *args)

    def _bin(self, bin):
        return bin



class BaseInstaller:

    def install(self, package):
        raise NotImplementedError

    def update(self, source, target):
        raise NotImplementedError

    def remove(self, package):
        raise NotImplementedError



from .prettify.elements.table import TableElement


class FreshTable(TableElement):
    """
    A fresh TableElement that appended itself to each of parents when it first gets written to at most once.

    parents is a sequence of objects providing an append_fresh_table(TableElement) method
    """

    def __init__(self, parent, name, is_array=False):
        TableElement.__init__(self, sub_elements=[])

        self._parent = parent
        self._name = name
        self._is_array = is_array

        # As long as this flag is false, setitem() operations will append the table header and this table
        # to the toml_file's elements
        self.__appended = False

    @property
    def name(self):
        return self._name

    @property
    def is_array(self):
        return self._is_array

    def _append_to_parent(self):
        """
        Causes this ephemeral table to be persisted on the TOMLFile.
        """

        if self.__appended:
            return

        if self._parent is not None:
            self._parent.append_fresh_table(self)

        self.__appended = True

    def __setitem__(self, key, value):
        TableElement.__setitem__(self, key, value)
        self._append_to_parent()



import json
import re

from poetry.config import Config

from .command import Command


TEMPLATE = """[settings]

[repositories]
"""

AUTH_TEMPLATE = """[http-basic]
"""


class ConfigCommand(Command):
    """
    Sets/Gets config options.

    config
        { key : Setting key. }
        { value?* : Setting value. }
        { --list : List configuration settings }
        { --unset : Unset configuration setting }
    """

    help = """This command allows you to edit the poetry config settings and repositories..

To add a repository:

    <comment>poetry repositories.foo https://bar.com/simple/</comment>

To remove a repository (repo is a short alias for repositories):

    <comment>poetry --unset repo.foo</comment>
"""

    def __init__(self):
        super().__init__()

        self._config = Config.create('config.toml')
        self._auth_config = Config.create('auth.toml')

    def initialize(self, i, o):
        super().initialize(i, o)

        # Create config file if it does not exist
        if not self._config.file.exists():
            self._config.file.parent.mkdir(parents=True, exist_ok=True)
            self._config.file.write_text(TEMPLATE)

        if not self._auth_config.file.exists():
            self._auth_config.file.parent.mkdir(parents=True, exist_ok=True)
            self._auth_config.file.write_text(AUTH_TEMPLATE)

    def handle(self):
        if self.option('list'):
            self._list_configuration(self._config.raw_content)

            return 0

        setting_key = self.argument('key')
        if not setting_key:
            return 0

        if self.argument('value') and self.option('unset'):
            raise RuntimeError('You can not combine a setting value with --unset')

        # show the value if no value is provided
        if not self.argument('value') and not self.option('unset'):
            m = re.match('^repos?(?:itories)?(?:\.(.+))?', self.argument('key'))
            if m:
                if not m.group(1):
                    value = {}
                    if self._config.setting('repositories') is not None:
                        value = self._config.setting('repositories')
                else:
                    repo = self._config.setting(f'repositories.{m.group(1)}')
                    if repo is None:
                        raise ValueError(
                            f'There is no {m.group(1)} repository defined'
                        )

                    value = repo

                self.line(str(value))

            return 0

        values = self.argument('value')

        boolean_validator = lambda val: val in {'true', 'false', '1', '0'}
        boolean_normalizer = lambda val: True if val in ['true', '1'] else False

        unique_config_values = {
            'settings.virtualenvs.create': (boolean_validator, boolean_normalizer)
        }

        if setting_key in unique_config_values:
            if self.option('unset'):
                return self._remove_single_value(setting_key)

            return self._handle_single_value(
                setting_key,
                unique_config_values[setting_key],
                values
            )

        # handle repositories
        m = re.match('^repos?(?:itories)?(?:\.(.+))?', self.argument('key'))
        if m:
            if not m.group(1):
                raise ValueError('You cannot remove the [repositories] section')

            if self.option('unset'):
                repo = self._config.setting(f'repositories.{m.group(1)}')
                if repo is None:
                    raise ValueError(f'There is no {m.group(1)} repository defined')

                self._config.remove_property(f'repositories.{m.group(1)}')

                return 0

            if len(values) == 1:
                url = values[0]

                self._config.add_property(f'repositories.{m.group(1)}.url', url)

                return 0

            raise ValueError(
                'You must pass the url. '
                'Example: poetry config repositories.foo https://bar.com'
            )

        # handle auth
        m = re.match('^(http-basic)\.(.+)', self.argument('key'))
        if m:
            if self.option('unset'):
                if not self._auth_config.setting(f'{m.group(1)}.{m.group(2)}'):
                    raise ValueError(
                        f'There is no {m.group(2)} {m.group(1)} defined'
                    )

                self._auth_config.remove_property(f'{m.group(1)}.{m.group(2)}')

                return 0

            if m.group(1) == 'http-basic':
                if len(values) == 1:
                    username = values[0]
                    # Only username, so we prompt for password
                    password = self.secret('Password:')
                elif len(values) != 2:
                    raise ValueError(f'Expected one or two arguments '
                                     f'(username, password), got {len(values)}')
                else:
                    username = values[0]
                    password = values[1]

                self._auth_config.add_property(
                    f'{m.group(1)}.{m.group(2)}', {
                        'username': username,
                        'password': password
                    }
                )

            return 0

        raise ValueError(f'Setting {self.argument("key")} does not exist')

    def _handle_single_value(self, key, callbacks, values):
        validator, normalizer = callbacks

        if len(values) > 1:
            raise RuntimeError('You can only pass one value.')

        value = values[0]
        if not validator(value):
            raise RuntimeError(
                f'"{value}" is an invalid value for {key}'
            )

        self._config.add_property(key, normalizer(value))

        return 0

    def _remove_single_value(self, key):
        self._config.remove_property(key)

        return 0

    def _list_configuration(self, contents, k=None):
        orig_k = k

        for key, value in contents.items():
            if k is None and key not in ['config', 'repositories']:
                continue

            if isinstance(value, dict) or key == 'repositories' and k is None:
                if k is None:
                    k = ''

                k += re.sub('^config\.', '', key + '.')
                self._list_configuration(value, k=k)
                k = orig_k

                continue

            if isinstance(value, list):
                value = [
                    json.dumps(val) if isinstance(val, list) else val
                    for val in value
                ]

                value = f'[{", ".join(value)}]'

            value = json.dumps(value)

            self.line(f'[<comment>{(k or "") + key}</comment>] <info>{value}</info>')



from pathlib import Path
from typing import Any

from .locations import CONFIG_DIR
from .utils.toml_file import TomlFile
from .utils.toml_file import TOMLFile


class Config:

    def __init__(self, file: TomlFile):
        self._file = file
        if not self._file.exists():
            self._raw_content = {}
            self._content = TOMLFile([])
        else:
            self._raw_content = file.read(raw=True)
            self._content = file.read()
        
    @property
    def name(self):
        return str(self._file.path)

    @property
    def file(self):
        return self._file

    @property
    def raw_content(self):
        return self._raw_content

    @property
    def content(self):
        return self._content

    def setting(self, setting_name: str) -> Any:
        """
        Retrieve a setting value.
        """
        keys = setting_name.split('.')

        config = self._raw_content
        for key in keys:
            if key not in config:
                return None

            config = config[key]

        return config

    def add_property(self, key, value):
        keys = key.split('.')

        config = self._content
        for i, key in enumerate(keys):
            if key not in config and i < len(keys) - 1:
                config[key] = {}

            if i == len(keys) - 1:
                config[key] = value
                break

            config = config[key]

        self.dump()

    def remove_property(self, key):
        keys = key.split('.')

        config = self._content
        for i, key in enumerate(keys):
            if key not in config:
                return

            if i == len(keys) - 1:
                del config[key]
                break

            config = config[key]

        self.dump()

    def dump(self):
        self._file.write(self._content)

    @classmethod
    def create(cls, file, base_dir=None) -> 'Config':
        if base_dir is None:
            base_dir = CONFIG_DIR

        file = TomlFile(Path(base_dir) / file)

        return cls(file)



import os
import tarfile

from collections import defaultdict
from copy import copy
from gzip import GzipFile
from io import BytesIO
from pathlib import Path
from posixpath import join as pjoin
from pprint import pformat
from typing import List

from poetry.packages import Dependency

from ..utils.helpers import normalize_file_permissions

from .builder import Builder


SETUP = """\
# -*- coding: utf-8 -*-
from distutils.core import setup

{before}
setup_kwargs = {{
    'name': {name!r},
    'version': {version!r},
    'description': {description!r},
    'long_description': {long_description!r},
    'author': {author!r},
    'author_email': {author_email!r},
    'url': {url!r},
    {extra}
}}
{after}

setup(**setup_kwargs)
"""


PKG_INFO = """\
Metadata-Version: 1.1
Name: {name}
Version: {version}
Summary: {summary}
Home-page: {home_page}
Author: {author}
Author-email: {author_email}
"""


class SdistBuilder(Builder):

    def build(self, target_dir: Path = None) -> Path:
        self._io.writeln(' - Building <info>sdist</info>')
        if target_dir is None:
            target_dir = self._path / 'dist'

        if not target_dir.exists():
            target_dir.mkdir(parents=True)

        target = target_dir / f'{self._package.pretty_name}' \
                              f'-{self._package.version}.tar.gz'
        gz = GzipFile(target.as_posix(), mode='wb')
        tar = tarfile.TarFile(target.as_posix(), mode='w', fileobj=gz,
                              format=tarfile.PAX_FORMAT)

        try:
            tar_dir = f'{self._package.pretty_name}-{self._package.version}'

            files_to_add = self.find_files_to_add(exclude_build=False)

            for relpath in files_to_add:
                path = self._path / relpath
                tar_info = tar.gettarinfo(
                    str(path),
                    arcname=pjoin(tar_dir, relpath)
                )
                tar_info = self.clean_tarinfo(tar_info)

                if tar_info.isreg():
                    with path.open('rb') as f:
                        tar.addfile(tar_info, f)
                else:
                    tar.addfile(tar_info)  # Symlinks & ?

            setup = self.build_setup()
            tar_info = tarfile.TarInfo(pjoin(tar_dir, 'setup.py'))
            tar_info.size = len(setup)
            tar.addfile(tar_info, BytesIO(setup))

            pkg_info = PKG_INFO.format(
                name=self._meta.name,
                version=self._meta.version,
                summary=self._meta.summary,
                home_page=self._meta.home_page,
                author=self._meta.author,
                author_email=self._meta.author_email,
            ).encode('utf-8')

            tar_info = tarfile.TarInfo(pjoin(tar_dir, 'PKG-INFO'))
            tar_info.size = len(pkg_info)
            tar.addfile(tar_info, BytesIO(pkg_info))
        finally:
            tar.close()
            gz.close()

        self._io.writeln(f' - Built <fg=cyan>{target.name}</>')

        return target

    def build_setup(self) -> bytes:
        before, extra, after = [], [], []

        # If we have a build script, use it
        if self._package.build:
            after += [
                f'from {self._package.build.split(".")[0]} import *',
                'build(setup_kwargs)'
            ]

        if self._module.is_package():
            packages, package_data = self.find_packages(
                self._module.path.as_posix()
            )
            before.append("packages = \\\n{}\n".format(pformat(sorted(packages))))
            before.append("package_data = \\\n{}\n".format(pformat(package_data)))
            extra.append("'packages': packages,")
            extra.append("'package_data': package_data,")
        else:
            extra.append("'py_modules': {!r},".format(self._module.name))

        dependencies, extras = self.convert_dependencies(self._package.requires)
        if dependencies:
            before.append("install_requires = \\\n{}\n".format(pformat(dependencies)))
            extra.append("'install_requires': install_requires,")

        if extras:
            before.append("extras_require = \\\n{}\n".format(pformat(extras)))
            extra.append("'extras_require': extras_require,")

        entry_points = self.convert_entry_points()
        if entry_points:
            before.append("entry_points = \\\n{}\n".format(pformat(entry_points)))
            extra.append("'entry_points': entry_points,")

        if self._package.python_versions != '*':
            python_requires = self._meta.requires_python

            extra.append("'python_requires': {!r},".format(python_requires))

        return SETUP.format(
            before='\n'.join(before),
            name=self._meta.name,
            version=self._meta.version,
            description=self._meta.summary,
            long_description=self._meta.description,
            author=self._meta.author,
            author_email=self._meta.author_email,
            url=self._meta.home_page,
            extra='\n    '.join(extra),
            after='\n'.join(after)
        ).encode('utf-8')

    @classmethod
    def find_packages(cls, path: str):
        """
        Discover subpackages and data.

        It also retrieve necessary files
        """
        pkgdir = os.path.normpath(path)
        pkg_name = os.path.basename(pkgdir)
        pkg_data = defaultdict(list)
        # Undocumented distutils feature:
        # the empty string matches all package names
        pkg_data[''].append('*')
        packages = [pkg_name]
        subpkg_paths = set()

        def find_nearest_pkg(rel_path):
            parts = rel_path.split(os.sep)
            for i in reversed(range(1, len(parts))):
                ancestor = '/'.join(parts[:i])
                if ancestor in subpkg_paths:
                    pkg = '.'.join([pkg_name] + parts[:i])
                    return pkg, '/'.join(parts[i:])

            # Relative to the top-level package
            return pkg_name, rel_path

        for path, dirnames, filenames in os.walk(pkgdir, topdown=True):
            if os.path.basename(path) == '__pycache__':
                continue

            from_top_level = os.path.relpath(path, pkgdir)
            if from_top_level == '.':
                continue

            is_subpkg = '__init__.py' in filenames
            if is_subpkg:
                subpkg_paths.add(from_top_level)
                parts = from_top_level.split(os.sep)
                packages.append('.'.join([pkg_name] + parts))
            else:
                pkg, from_nearest_pkg = find_nearest_pkg(from_top_level)
                pkg_data[pkg].append(pjoin(from_nearest_pkg, '*'))

        # Sort values in pkg_data
        pkg_data = {k: sorted(v) for (k, v) in pkg_data.items()}

        return sorted(packages), pkg_data

    @classmethod
    def convert_dependencies(cls,
                             dependencies: List[Dependency]):
        main = []
        extras = []

        for dependency in dependencies:
            requirement = dependency.to_pep_508()

            if ';' in requirement:
                extras.append(requirement)
            else:
                main.append(requirement)

        return main, extras

    @classmethod
    def clean_tarinfo(cls, tar_info):
        """
        Clean metadata from a TarInfo object to make it more reproducible.

            - Set uid & gid to 0
            - Set uname and gname to ""
            - Normalise permissions to 644 or 755
            - Set mtime if not None
        """
        ti = copy(tar_info)
        ti.uid = 0
        ti.gid = 0
        ti.uname = ''
        ti.gname = ''
        ti.mode = normalize_file_permissions(ti.mode)
        
        return ti



import re

from .constraint import Constraint


class WilcardConstraint(Constraint):

    def __init__(self, constraint: str):
        m = re.match(
            '^(!=|==)?v?(\d+)(?:\.(\d+))?(?:\.(\d+))?(?:\.[xX*])+$',
            constraint
        )
        if not m:
            raise ValueError('Invalid value for wildcard constraint')

        if not m.group(1):
            operator = '=='
        else:
            operator = m.group(1)

        super().__init__(
            operator,
            '.'.join([g if g else '*' for g in m.groups()[1:]])
        )

        if m.group(4):
            position = 2
        elif m.group(3):
            position = 1
        else:
            position = 0

        from ..version_parser import VersionParser

        parser = VersionParser()
        groups = m.groups()[1:]
        low_version = parser._manipulate_version_string(
            groups, position
        )
        high_version = parser._manipulate_version_string(
            groups, position, 1
        )

        if operator == '!=':
            if low_version == '0.0.0.0':
                self._constraint = Constraint('>=', high_version)
            else:
                self._constraint = parser.parse_constraints(
                    f'<{low_version} || >={high_version}'
                )
        else:
            if low_version == '0.0.0.0':
                self._constraint = Constraint('<', high_version)
            else:
                self._constraint = parser.parse_constraints(
                    f'>={low_version},<{high_version}'
                )

    @property
    def supported_operators(self):
        return ['!=', '==']

    @property
    def constraint(self):
        return self._constraint

    def matches(self, provider) -> bool:
        if isinstance(provider, self.__class__):
            return self._constraint.matches(provider.constraint)

        return provider.matches(self._constraint)

    def __str__(self):
        op = ''
        if self.string_operator == '!=':
            op = '!= '

        return '{}{}'.format(
            op,
            self._version
        )



from typing import Type

from .layout import Layout
from .standard import StandardLayout


_LAYOUTS = {
    'standard': StandardLayout
}


def layout(name: str) -> Type[Layout]:
    if name not in _LAYOUTS:
        raise ValueError('Invalid layout')

    return _LAYOUTS[name]



import os
import re

from collections import defaultdict
from pathlib import Path

from poetry.semver.constraints import Constraint
from poetry.semver.constraints import MultiConstraint
from poetry.semver.version_parser import VersionParser
from poetry.vcs import get_vcs

from ..metadata import Metadata
from ..utils.module import Module


AUTHOR_REGEX = re.compile('(?u)^(?P<name>[- .,\w\d\'’"()]+) <(?P<email>.+?)>$')


class Builder:

    AVAILABLE_PYTHONS = {
        '2',
        '2.7',
        '3',
        '3.4', '3.5', '3.6', '3.7'
    }

    def __init__(self, poetry, venv, io):
        self._poetry = poetry
        self._venv = venv
        self._io = io
        self._package = poetry.package
        self._path = poetry.file.parent
        self._module = Module(
            self._package.name, self._path.as_posix()
        )
        self._meta = Metadata.from_package(self._package)

    def build(self):
        raise NotImplementedError()

    def find_excluded_files(self) -> list:
        # Checking VCS
        vcs = get_vcs(self._path)
        if not vcs:
            return []

        ignored = vcs.get_ignored_files()
        result = []
        for file in ignored:
            try:
                file = Path(file).absolute().relative_to(self._path)
            except ValueError:
                # Should only happen in tests
                continue

            result.append(file)

        return result

    def find_files_to_add(self, exclude_build=True) -> list:
        """
        Finds all files to add to the tarball

        TODO: Support explicit include/exclude
        """
        excluded = self.find_excluded_files()
        src = self._module.path
        to_add = []

        for root, dirs, files in os.walk(src.as_posix()):
            root = Path(root)
            if root.name == '__pycache__':
                continue

            for file in files:
                file = root / file
                file = file.relative_to(self._path)

                if file in excluded:
                    continue

                if file.suffix == '.pyc':
                    continue

                self._io.writeln(
                    f' - Adding: <comment>{str(file)}</comment>',
                    verbosity=self._io.VERBOSITY_VERY_VERBOSE
                )
                to_add.append(file)

        # Include project files
        self._io.writeln(
            f' - Adding: <comment>pyproject.toml</comment>',
            verbosity=self._io.VERBOSITY_VERY_VERBOSE
        )
        to_add.append(Path('pyproject.toml'))

        # If a README is specificed we need to include it
        # to avoid errors
        if 'readme' in self._poetry.config:
            readme = self._path / self._poetry.config['readme']
            if readme.exists():
                self._io.writeln(
                    f' - Adding: <comment>{readme.relative_to(self._path)}</comment>',
                    verbosity=self._io.VERBOSITY_VERY_VERBOSE
                )
                to_add.append(readme.relative_to(self._path))

        # If a build script is specified and explicitely required
        # we add it to the list of files
        if self._package.build and not exclude_build:
            to_add.append(Path(self._package.build))

        return sorted(to_add)

    def convert_entry_points(self) -> dict:
        result = defaultdict(list)

        # Scripts -> Entry points
        for name, ep in self._poetry.config.get('scripts', {}).items():
            result['console_scripts'].append(f'{name} = {ep}')

        # Plugins -> entry points
        for groupname, group in self._poetry.config.get('plugins', {}).items():
            for name, ep in sorted(group.items()):
                result[groupname].append(f'{name} = {ep}')

        return dict(result)

    @classmethod
    def convert_author(cls, author) -> dict:
        m = AUTHOR_REGEX.match(author)

        name = m.group('name')
        email = m.group('email')

        return {
            'name': name,
            'email': email
        }

    def get_classifers(self):
        classifiers = []

        # Automatically set python classifiers
        parser = VersionParser()
        if self._package.python_versions == '*':
            python_constraint = parser.parse_constraints('~2.7 || ^3.4')
        else:
            python_constraint = self._package.python_constraint

        for version in sorted(self.AVAILABLE_PYTHONS):
            if python_constraint.matches(Constraint('=', version)):
                classifiers.append(f'Programming Language :: Python :: {version}')

        return classifiers

    def convert_python_version(self):
        constraint = self._package.python_constraint
        if isinstance(constraint, MultiConstraint):
            python_requires = ','.join(
                [str(c).replace(' ', '') for c in constraint.constraints]
            )
        else:
            python_requires = str(constraint).replace(' ', '')

        return python_requires



import poetry.packages

from poetry.semver.constraints import Constraint
from poetry.semver.constraints import EmptyConstraint
from poetry.semver.constraints import MultiConstraint
from poetry.semver.constraints.base_constraint import BaseConstraint
from poetry.semver.version_parser import VersionParser

from .constraints.platform_constraint import PlatformConstraint


class Dependency:

    def __init__(self,
                 name: str,
                 constraint: str,
                 optional: bool = False,
                 category: str = 'main',
                 allows_prereleases: bool = False):
        self._name = name.lower()
        self._pretty_name = name
        self._parser = VersionParser()

        try:
            if not isinstance(constraint, BaseConstraint):
                self._constraint = self._parser.parse_constraints(constraint)
            else:
                self._constraint = constraint
        except ValueError:
            self._constraint = self._parser.parse_constraints('*')

        self._pretty_constraint = constraint
        self._optional = optional
        self._category = category
        self._allows_prereleases = allows_prereleases

        self._python_versions = '*'
        self._python_constraint = self._parser.parse_constraints('*')
        self._platform = '*'
        self._platform_constraint = EmptyConstraint()

        self._extras = []

    @property
    def name(self):
        return self._name

    @property
    def constraint(self):
        return self._constraint
    
    @property
    def pretty_constraint(self):
        return self._pretty_constraint

    @property
    def pretty_name(self):
        return self._pretty_name

    @property
    def category(self):
        return self._category

    @property
    def python_versions(self):
        return self._python_versions

    @python_versions.setter
    def python_versions(self, value: str):
        self._python_versions = value
        self._python_constraint = self._parser.parse_constraints(value)

    @property
    def python_constraint(self):
        return self._python_constraint

    @property
    def platform(self) -> str:
        return self._platform

    @platform.setter
    def platform(self, value: str):
        self._platform = value
        self._platform_constraint = PlatformConstraint.parse(value)

    @property
    def platform_constraint(self):
        return self._platform_constraint

    @property
    def extras(self) -> list:
        return self._extras

    def allows_prereleases(self):
        return self._allows_prereleases

    def is_optional(self):
        return self._optional

    def is_vcs(self):
        return False

    def accepts(self, package: 'poetry.packages.Package') -> bool:
        """
        Determines if the given package matches this dependency.
        """
        return (
            self._name == package.name
            and self._constraint.matches(Constraint('=', package.version))
            and (not package.is_prerelease() or self.allows_prereleases())
        )

    def to_pep_508(self) -> str:
        requirement = f'{self.pretty_name}'

        if isinstance(self.constraint, MultiConstraint):
            requirement += ' ({})'.format(','.join(
                [str(c).replace(' ', '') for c in self.constraint.constraints]
            ))
        else:
            requirement += ' ({})'.format(str(self.constraint).replace(' ', ''))

        # Markers
        markers = []

        # Python marker
        if self.python_versions != '*':
            python_constraint = self.python_constraint

            markers.append(self._create_nested_marker('python_version', python_constraint))

        if markers:
            requirement += f'; {" and ".join(markers)}'

        return requirement

    @classmethod
    def from_pep_508(cls):
        return

    def _create_nested_marker(self, name, constraint):
        if isinstance(constraint, MultiConstraint):
            parts = []
            for c in constraint.constraints:
                parts.append(self._create_nested_marker(name, c))

            glue = ' and '
            if constraint.is_disjunctive():
                parts = [f'({part})' for part in parts]
                glue = ' or '

            marker = glue.join(parts)
        else:
            marker = f'{name}{constraint.string_operator}"{constraint.version}"'

        return marker

    def activate(self):
        """
        Set the dependency as mandatory.
        """
        self._optional = False

    def deactivate(self):
        """
        Set the dependency as optional.
        """
        self._optional = True

    def __eq__(self, other):
        if not isinstance(other, Dependency):
            return NotImplemented

        return self._name == other.name and self._constraint == other.constraint

    def __hash__(self):
        return hash((self._name, self._pretty_constraint))

    def __str__(self):
        return f'{self._pretty_name} ({self._pretty_constraint})'

    def __repr__(self):
        return f'<Dependency {str(self)}>'



from poetry.masonry.publishing.publisher import Publisher

from .command import Command


class PublishCommand(Command):
    """
    Publishes a package to a remote repository.

    publish
        { --r|repository= : The repository to publish the package to. }
        { --no-build : Do not build the package before publishing. }
    """

    help = """The publish command builds and uploads the package to a remote repository.
    
By default, it will upload to PyPI but if you pass the --repository option it will
upload to it instead.

The --repository option should match the name of a configured repository using
the config command.
"""

    def handle(self):
        # Building package first, unless told otherwise
        if not self.option('no-build'):
            self.call('build')

        self.line('')

        publisher = Publisher(self.poetry, self.output)
        publisher.publish(self.option('repository'))



def flat_map(iter, callable):
    if not isinstance(iter, (list, tuple)):
        yield callable(iter)
    else:
        for v in iter:
            for i in flat_map(v, callable):
                yield i



import sys

from typing import List

from poetry.packages import Dependency
from poetry.packages import Locker
from poetry.packages import Package
from poetry.packages.constraints.platform_constraint import PlatformConstraint
from poetry.puzzle import Solver
from poetry.puzzle.operations import Install
from poetry.puzzle.operations import Uninstall
from poetry.puzzle.operations import Update
from poetry.puzzle.operations.operation import Operation
from poetry.repositories import Pool
from poetry.repositories import Repository
from poetry.repositories.installed_repository import InstalledRepository
from poetry.semver.constraints import Constraint
from poetry.semver.version_parser import VersionParser

from .base_installer import BaseInstaller
from .pip_installer import PipInstaller


class Installer:

    def __init__(self,
                 io,
                 venv,
                 package: Package,
                 locker: Locker,
                 pool: Pool):
        self._io = io
        self._venv = venv
        self._package = package
        self._locker = locker
        self._pool = pool

        self._dry_run = False
        self._update = False
        self._verbose = False
        self._write_lock = True
        self._dev_mode = True
        self._execute_operations = True

        self._whitelist = {}

        self._extras = []

        self._installer = self._get_installer()

    @property
    def installer(self):
        return self._installer

    def run(self):
        # Force update if there is no lock file present
        if not self._update and not self._locker.is_locked():
            self._update = True

        if self.is_dry_run():
            self.verbose(True)
            self._write_lock = False
            self._execute_operations = False

        local_repo = Repository()
        self._do_install(local_repo)

        return 0

    def dry_run(self, dry_run=True) -> 'Installer':
        self._dry_run = dry_run

        return self

    def is_dry_run(self) -> bool:
        return self._dry_run

    def verbose(self, verbose=True) -> 'Installer':
        self._verbose = verbose

        return self

    def is_verbose(self) -> bool:
        return self._verbose

    def dev_mode(self, dev_mode=True) -> 'Installer':
        self._dev_mode = dev_mode

        return self

    def is_dev_mode(self) -> bool:
        return self._dev_mode

    def update(self, update=True) -> 'Installer':
        self._update = update

        return self

    def is_updating(self) -> bool:
        return self._update

    def execute_operations(self, execute=True) -> 'Installer':
        self._execute_operations = execute

        return self

    def whitelist(self, packages: dict) -> 'Installer':
        self._whitelist = packages

        return self

    def extras(self, extras: list) -> 'Installer':
        self._extras = extras

        return self

    def _do_install(self, local_repo):
        locked_repository = Repository()
        # initialize locked repo if we are installing from lock
        if not self._update or (self._update and self._locker.is_locked()):
            locked_repository = self._locker.locked_repository(True)

        if self._update:
            # Checking extras
            for extra in self._extras:
                if extra not in self._package.extras:
                    raise ValueError(f'Extra [{extra}] is not specified.')

            self._io.writeln('<info>Updating dependencies</>')
            fixed = []

            # If the whitelist is enabled, packages not in it are fixed
            # to the version specified in the lock
            if self._whitelist:
                # collect packages to fixate from root requirements
                candidates = []
                for package in locked_repository.packages:
                    candidates.append(package)

                # fix them to the version in lock if they are not updateable
                for candidate in candidates:
                    to_fix = True
                    for require in self._whitelist.keys():
                        if require == candidate.name:
                            to_fix = False

                    if to_fix:
                        fixed.append(
                            Dependency(candidate.name, candidate.version)
                        )

            solver = Solver(
                self._package,
                self._pool,
                locked_repository,
                self._io
            )

            request = self._package.requires
            request += self._package.dev_requires

            ops = solver.solve(request, fixed=fixed)
        else:
            self._io.writeln('<info>Installing dependencies from lock file</>')
            if not self._locker.is_fresh():
                self._io.writeln(
                    '<warning>'
                    'Warning: The lock file is not up to date with '
                    'the latest changes in pyproject.toml. '
                    'You may be getting outdated dependencies. '
                    'Run update to update them.'
                    '</warning>'
                )

            for extra in self._extras:
                if extra not in self._locker.lock_data.get('extras', {}):
                    raise ValueError(f'Extra [{extra}] is not specified.')

            # If we are installing from lock
            # Filter the operations by comparing it with what is
            # currently installed
            ops = self._get_operations_from_lock(locked_repository)

        self._populate_local_repo(local_repo, ops, locked_repository)

        # We need to filter operations so that packages
        # not compatible with the current system,
        # or optional and not requested, are dropped
        self._filter_operations(ops, local_repo)

        self._io.new_line()

        # Execute operations
        if not ops and (self._execute_operations or self._dry_run):
            self._io.writeln('Nothing to install or update')

        if ops and (self._execute_operations or self._dry_run):
            installs = []
            updates = []
            uninstalls = []
            skipped = []
            for op in ops:
                if op.skipped:
                    skipped.append(op)
                    continue

                if op.job_type == 'install':
                    installs.append(
                        f'{op.package.pretty_name}'
                        f':{op.package.full_pretty_version}'
                    )
                elif op.job_type == 'update':
                    updates.append(
                        f'{op.target_package.pretty_name}'
                        f':{op.target_package.full_pretty_version}'
                    )
                elif op.job_type == 'uninstall':
                    uninstalls.append(
                        f'{op.package.pretty_name}'
                    )

            self._io.new_line()
            self._io.writeln(
                'Package operations: '
                f'<info>{len(installs)}</> install{"" if len(installs) == 1 else "s"}, '
                f'<info>{len(updates)}</> update{"" if len(updates) == 1 else "s"}, '
                f'<info>{len(uninstalls)}</> removal{"" if len(uninstalls) == 1 else "s"}'
                f'{", <info>{}</> skipped".format(len(skipped)) if skipped and self.is_verbose() else ""}'
            )
            self._io.new_line()

        # Writing lock before installing
        if self._update and self._write_lock:
            updated_lock = self._locker.set_lock_data(
                self._package,
                local_repo.packages
            )

            if updated_lock:
                self._io.writeln('<info>Writing lock file</>')
                self._io.writeln('')

        for op in ops:
            self._execute(op)

    def _execute(self, operation: Operation) -> None:
        """
        Execute a given operation.
        """
        method = operation.job_type

        getattr(self, f'_execute_{method}')(operation)

    def _execute_install(self, operation: Install) -> None:
        if operation.skipped:
            if self.is_verbose() and (self._execute_operations or self.is_dry_run()):
                self._io.writeln(
                    f'  - Skipping <info>{operation.package.pretty_name}</> '
                    f'(<comment>{operation.package.full_pretty_version}</>) '
                    f'{operation.skip_reason}')

            return

        if self._execute_operations or self.is_dry_run():
            self._io.writeln(
                f'  - Installing <info>{operation.package.pretty_name}</> '
                f'(<comment>{operation.package.full_pretty_version}</>)'
            )

        if not self._execute_operations:
            return

        self._installer.install(operation.package)

    def _execute_update(self, operation: Update) -> None:
        source = operation.initial_package
        target = operation.target_package

        if operation.skipped:
            if self.is_verbose() and (self._execute_operations or self.is_dry_run()):
                self._io.writeln(
                    f'  - Skipping <info>{target.pretty_name}</> '
                    f'(<comment>{target.full_pretty_version}</>) '
                    f'{operation.skip_reason}')

            return

        if self._execute_operations or self.is_dry_run():
            self._io.writeln(
                f'  - Updating <info>{target.pretty_name}</> '
                f'(<comment>{source.pretty_version}</>'
                f' -> <comment>{target.pretty_version}</>)'
            )

        if not self._execute_operations:
            return

        self._installer.update(source, target)

    def _execute_uninstall(self, operation: Uninstall) -> None:
        if self._execute_operations or self.is_dry_run():
            self._io.writeln(
                f'  - Removing <info>{operation.package.pretty_name}</> '
                f'(<comment>{operation.package.full_pretty_version}</>)'
            )

        if not self._execute_operations:
            return

        self._installer.remove(operation.package)

    def _populate_local_repo(self, local_repo, ops, locked_repository):
        # Add all locked packages from the lock and go from there
        for package in locked_repository.packages:
            local_repo.add_package(package)

        # Now, walk through all operations and add/remove/update accordingly
        for op in ops:
            if isinstance(op, Update):
                package = op.target_package
            else:
                package = op.package

            acted_on = False
            for pkg in local_repo.packages:
                if pkg.name == package.name:
                    # The package we operate on is in the local repo
                    if op.job_type == 'update':
                        if pkg.version == package.version:
                            break

                        local_repo.remove_package(pkg)
                        local_repo.add_package(op.target_package)
                    elif op.job_type == 'uninstall':
                        local_repo.remove_package(op.package)

                    acted_on = True

            if not acted_on:
                local_repo.add_package(package)

    def _get_operations_from_lock(self,
                                  locked_repository: Repository
                                  ) -> List[Operation]:
        installed_repo = InstalledRepository.load(self._venv)
        ops = []

        extra_packages = [
            p.name
            for p in self._get_extra_packages(locked_repository)
        ]
        for locked in locked_repository.packages:
            is_installed = False
            for installed in installed_repo.packages:
                if locked.name == installed.name:
                    is_installed = True
                    if locked.category == 'dev' and not self.is_dev_mode():
                        ops.append(Uninstall(locked))
                    elif locked.optional and locked.name not in extra_packages:
                        # Installed but optional and not requested in extras
                        ops.append(Uninstall(locked))
                    elif locked.version != installed.version:
                        ops.append(Update(
                            installed, locked
                        ))

            if not is_installed:
                # If it's optional and not in required extras
                # we do not install
                if locked.optional and locked.name not in extra_packages:
                    continue

                ops.append(Install(locked))

        return ops

    def _filter_operations(self, ops: List[Operation], repo: Repository) -> None:
        extra_packages = [p.name for p in
                          self._get_extra_packages(repo)]
        for op in ops:
            if isinstance(op, Update):
                package = op.target_package
            else:
                package = op.package

            if op.job_type == 'uninstall':
                continue

            parser = VersionParser()
            python = '.'.join([str(i) for i in self._venv.version_info[:3]])
            if 'python' in package.requirements:
                python_constraint = parser.parse_constraints(
                    package.requirements['python']
                )
                if not python_constraint.matches(Constraint('=', python)):
                    # Incompatible python versions
                    op.skip('Not needed for the current python version')
                    continue

            if 'platform' in package.requirements:
                platform_constraint = PlatformConstraint.parse(
                    package.requirements['platform']
                )
                if not platform_constraint.matches(
                        PlatformConstraint('=', sys.platform)
                ):
                    # Incompatible systems
                    op.skip('Not needed for the current platform')
                    continue

            if self._update:
                extras = {}
                for extra, deps in self._package.extras.items():
                    extras[extra] = [dep.name for dep in deps]
            else:
                extras = {}
                for extra, deps in self._locker.lock_data.get('extras', {}).items():
                    extras[extra] = [dep.lower() for dep in deps]

            # If a package is optional and not requested
            # in any extra we skip it
            if package.optional:
                if package.name not in extra_packages:
                    op.skip('Not required')

    def _get_extra_packages(self, repo):
        """
        Returns all packages required by extras.

        Maybe we just let the solver handle it?
        """
        if self._update:
            extras = {
                k: [d.name for d in v]
                for k, v in self._package.extras.items()
            }
        else:
            extras = self._locker.lock_data.get('extras', {})

        extra_packages = []
        for extra_name, packages in extras.items():
            if extra_name not in self._extras:
                continue

            extra_packages += [Dependency(p, '*') for p in packages]

        def _extra_packages(packages):
            pkgs = []
            for package in packages:
                for pkg in repo.packages:
                    if pkg.name == package.name:
                        pkgs.append(package)
                        pkgs += _extra_packages(pkg.requires)

                        break

            return pkgs

        return _extra_packages(extra_packages)

    def _get_installer(self) -> BaseInstaller:
        return PipInstaller(self._venv, self._io)



from functools import cmp_to_key

from .comparison import less_than
from .constraints import Constraint
from .helpers import normalize_version
from .version_parser import VersionParser

SORT_ASC = 1
SORT_DESC = -1

_parser = VersionParser()


def statisfies(version, constraints):
    """
    Determine if given version satisfies given constraints.

    :type version: str
    :type constraints: str

    :rtype: bool
    """
    provider = Constraint('==', normalize_version(version))
    constraints = _parser.parse_constraints(constraints)

    return constraints.matches(provider)


def satisfied_by(versions, constraints):
    """
    Return all versions that satisfy given constraints.

    :type versions: List[str]
    :type constraints: str

    :rtype: List[str]
    """
    return [version for version in versions if statisfies(version, constraints)]


def sort(versions):
    return _sort(versions, SORT_ASC)


def rsort(versions):
    return _sort(versions, SORT_DESC)


def _sort(versions, direction):
    normalized = [
        (i, normalize_version(version))
        for i, version in enumerate(versions)
    ]
    normalized.sort(
        key=cmp_to_key(
            lambda x, y:
                0 if x[1] == y[1]
                else -direction * int(less_than(x[1], y[1]) or -1)
        )
    )

    return [versions[i] for i, _ in normalized]



from pip import __version__

from poetry.version import parse

if parse(__version__) >= parse('9.0.2'):
    from pip._internal.req import InstallRequirement
    from pip._internal.utils.appdirs import user_cache_dir, user_config_dir
else:
    from pip.req import InstallRequirement
    from pip.utils.appdirs import user_cache_dir, user_config_dir



from typing import Any
from typing import List
from typing import Union

from .contracts import SpecificationProvider
from .contracts import UI
from .dependency_graph import DependencyGraph
from .resolution import Resolution


class Resolver:

    def __init__(self,
                 specification_provider: SpecificationProvider,
                 resolver_ui: UI):
        self._specification_provider = specification_provider
        self._resolver_ui = resolver_ui

    @property
    def specification_provider(self) -> SpecificationProvider:
        return self._specification_provider

    @property
    def ui(self) -> UI:
        return self._resolver_ui

    def resolve(self,
                requested: List[Any],
                base: Union[DependencyGraph, None] = None) -> DependencyGraph:
        if base is None:
            base = DependencyGraph()

        return Resolution(
            self._specification_provider,
            self._resolver_ui,
            requested,
            base
        ).resolve()



from . import abstracttable, common, factory
from .errors import InvalidElementError
from .common import Element
from .metadata import CommentElement, NewlineElement, WhitespaceElement


class TableElement(abstracttable.AbstractTable):
    """
    An Element containing an unnamed top-level table.

    Implements dict-like interface.

    Assumes input sub_elements are correct.

    Raises InvalidElementError on duplicate keys.
    """

    def __init__(self, sub_elements):
        abstracttable.AbstractTable.__init__(self, sub_elements)

        self._check_for_duplicate_keys()

    def _check_for_duplicate_keys(self):
        if len(set(self.keys())) < len(self.keys()):
            raise InvalidElementError('Duplicate keys found')

    def __setitem__(self, key, value):
        if key in self:
            self._update(key, value)
        else:
            self._insert(key, value)

    def _update(self, key, value):
        _, value_i = self._find_key_and_value(key)
        self._sub_elements[value_i] = value if isinstance(value, Element) else factory.create_element(value)

    def _find_insertion_index(self):
        """
        Returns the self.sub_elements index in which new entries should be inserted.
        """

        non_metadata_elements = tuple(self._enumerate_non_metadata_sub_elements())

        if not non_metadata_elements:
            return 0

        last_entry_i = non_metadata_elements[-1][0]
        following_newline_i = self._find_following_line_terminator(last_entry_i)

        return following_newline_i + 1

    def _detect_indentation_size(self):
        """
        Detects the level of indentation used in this table.
        """

        def lines():
            # Returns a sequence of sequences of elements belonging to each line
            start = 0
            for i, element in enumerate(self.elements):
                if isinstance(element, (CommentElement, NewlineElement)):
                    yield self.elements[start:i+1]
                    start = i+1

        def indentation(line):
            # Counts the number of whitespace tokens at the beginning of this line
            try:
                first_non_whitespace_i = next(i for (i, e) in enumerate(line) if not isinstance(e, WhitespaceElement))
                return sum(space.length for space in line[:first_non_whitespace_i])
            except StopIteration:
                return 0

        def is_empty_line(line):
            return all(e.type == common.TYPE_METADATA for e in line)

        try:
            return min(indentation(line) for line in lines() if len(line) > 1 and not is_empty_line(line))
        except ValueError:  # Raised by ValueError when no matching lines found
            return 0

    def _insert(self, key, value):

        value_element = value if isinstance(value, Element) else factory.create_element(value)

        indentation_size = self._detect_indentation_size()
        indentation = [factory.create_whitespace_element(self._detect_indentation_size())] if indentation_size else []

        inserted_elements = indentation + [
            factory.create_string_element(key, bare_allowed=True),
            factory.create_whitespace_element(),
            factory.create_operator_element('='),
            factory.create_whitespace_element(),
            value_element,
            factory.create_newline_element(),
        ]
        
        insertion_index = self._find_insertion_index()
        
        self._sub_elements = \
            self.sub_elements[:insertion_index] + inserted_elements + self.sub_elements[insertion_index:]

    def __delitem__(self, key):
        begin, _ = self._find_key_and_value(key)
        preceding_newline = self._find_preceding_newline(begin)
        if preceding_newline >= 0:
            begin = preceding_newline
        end = self._find_following_newline(begin)
        if end < 0:
            end = len(tuple(self._sub_elements))
        self._sub_elements = self.sub_elements[:begin] + self.sub_elements[end:]

    @property
    def value(self):
        return self

    def __eq__(self, other):
        return self.primitive_value == other

    def __iter__(self):
        return iter(self.keys())

    def __str__(self):
        return str(self.primitive_value)



import operator
import re

from poetry.semver.constraints import EmptyConstraint
from poetry.semver.constraints import MultiConstraint
from poetry.semver.constraints.base_constraint import BaseConstraint


class PlatformConstraint(BaseConstraint):

    OP_EQ = operator.eq
    OP_NE = operator.ne

    _trans_op_str = {
        '=': OP_EQ,
        '==': OP_EQ,
        '!=': OP_NE
    }

    _trans_op_int = {
        OP_EQ: '==',
        OP_NE: '!='
    }

    def __init__(self, operator, platform):
        if operator not in self._trans_op_str:
            raise ValueError(
                f'Invalid operator "{operator}" given, '
                f'expected one of: {", ".join(self.supported_operators)}'
            )

        self._operator = self._trans_op_str[operator]
        self._string_operator = operator
        self._platform = platform

    @property
    def supported_operators(self) -> list:
        return list(self._trans_op_str.keys())

    @property
    def operator(self):
        return self._operator

    @property
    def string_operator(self):
        return self._string_operator

    @property
    def platform(self) -> str:
        return self._platform

    def matches(self, provider):
        if not isinstance(provider, (PlatformConstraint, EmptyConstraint)):
            raise ValueError(
                'Platform constraints can only be compared with each other'
            )

        if isinstance(provider, EmptyConstraint):
            return True

        is_equal_op = self.OP_EQ is self._operator
        is_non_equal_op = self.OP_NE is self._operator
        is_provider_equal_op = self.OP_EQ is provider.operator
        is_provider_non_equal_op = self.OP_NE is provider.operator

        if (
            is_equal_op and is_provider_equal_op
            or is_non_equal_op and is_provider_non_equal_op
        ):
            return self._platform == provider.platform

        if (
            is_equal_op and is_provider_non_equal_op
            or is_non_equal_op and is_provider_equal_op
        ):
            return self._platform != provider.platform

        return False

    @classmethod
    def parse(cls, constraints):
        """
        Parses a constraint string into
        MultiConstraint and/or PlatformConstraint objects.
        """
        pretty_constraint = constraints

        or_constraints = re.split('\s*\|\|?\s*', constraints.strip())
        or_groups = []
        for constraints in or_constraints:
            and_constraints = re.split(
                '(?<!^)(?<![ ,]) *(?<!-)[, ](?!-) *(?!,|$)',
                constraints
            )
            if len(and_constraints) > 1:
                constraint_objects = []
                for constraint in and_constraints:
                    for parsed_constraint in cls._parse_constraint(constraint):
                        constraint_objects.append(parsed_constraint)
            else:
                constraint_objects = cls._parse_constraint(and_constraints[0])

            if len(constraint_objects) == 1:
                constraint = constraint_objects[0]
            else:
                constraint = MultiConstraint(constraint_objects)

            or_groups.append(constraint)

        if len(or_groups) == 1:
            constraint = or_groups[0]
        else:
            constraint = MultiConstraint(or_groups, False)

        constraint.pretty_string = pretty_constraint

        return constraint

    @classmethod
    def _parse_constraint(cls, constraint):
        m = re.match('(?i)^v?[xX*](\.[xX*])*$', constraint)
        if m:
            return EmptyConstraint(),

        # Basic Comparators
        m = re.match('^(!=|==?)?\s*(.*)', constraint)
        if m:
            return PlatformConstraint(m.group(1) or '=', m.group(2)),

        raise ValueError(
            'Could not parse platform constraint: {}'.format(constraint)
        )

    def __str__(self):
        op = self._trans_op_int[self._operator]
        if op == '==':
            op = ''
        else:
            op = op + ' '

        return '{}{}'.format(
            op,
            self._platform
        )

    def __repr__(self):
        return '<PlatformConstraint \'{}\'>'.format(str(self))



import os
import tarfile

import poetry

from contextlib import contextmanager
from tempfile import TemporaryDirectory
from types import SimpleNamespace

from .builder import Builder
from .sdist import SdistBuilder
from .wheel import WheelBuilder


class CompleteBuilder(Builder):

    def build(self):
        # We start by building the tarball
        # We will use it to build the wheel
        sdist_builder = SdistBuilder(self._poetry, self._venv, self._io)
        sdist_file = sdist_builder.build()
        sdist_info = SimpleNamespace(builder=sdist_builder, file=sdist_file)

        self._io.writeln('')

        dist_dir = self._path / 'dist'
        with self.unpacked_tarball(sdist_file) as tmpdir:
            wheel_info = WheelBuilder.make_in(
                poetry.Poetry.create(tmpdir), self._venv, self._io, dist_dir,
                original=self._poetry
            )

        return SimpleNamespace(wheel=wheel_info, sdist=sdist_info)

    @classmethod
    @contextmanager
    def unpacked_tarball(cls, path):
        tf = tarfile.open(str(path))

        with TemporaryDirectory() as tmpdir:
            tf.extractall(tmpdir)
            files = os.listdir(tmpdir)

            assert len(files) == 1, files

            yield os.path.join(tmpdir, files[0])



from .exceptions import CircularDependencyError
from .graph.log import Log


class DependencyGraph:

    def __init__(self):
        self._vertices = {}
        self._log = Log()

    @property
    def vertices(self):
        return self._vertices

    @property
    def log(self):
        return self._log

    def tag(self, tag):
        return self._log.tag(self, tag)

    def rewind_to(self, tag):
        return self._log.rewind_to(self, tag)

    def add_child_vertex(self, name, payload, parent_names, requirement):
        root = True

        try:
            parent_names.index(None)
        except ValueError:
            root = False

        parent_names = [n for n in parent_names if n is not None]
        vertex = self.add_vertex(name, payload, root)
        if root:
            vertex.explicit_requirements.append(requirement)

        for parent_name in parent_names:
            parent_vertex = self.vertex_named(parent_name)
            self.add_edge(parent_vertex, vertex, requirement)

        return vertex

    def add_vertex(self, name, payload, root=False):
        return self._log.add_vertex(self, name, payload, root)

    def detach_vertex_named(self, name):
        return self._log.detach_vertex_named(self, name)

    def vertex_named(self, name):
        return self.vertices.get(name)

    def root_vertex_named(self, name):
        vertex = self.vertex_named(name)
        if vertex and vertex.root:
            return vertex

    def add_edge(self, origin, destination, requirement):
        if destination.has_path_to(origin):
            raise CircularDependencyError([origin, destination])

        return self.add_edge_no_circular(origin, destination, requirement)

    def add_edge_no_circular(self, origin, destination, requirement):
        self._log.add_edge_no_circular(
            self,
            origin.name, destination.name,
            requirement
        )

    def delete_edge(self, edge):
        return self._log.delete_edge(
            self,
            edge.origin.name,
            edge.destination.name,
            edge.requirement
        )

    def set_payload(self, name, payload):
        return self._log.set_payload(self, name, payload)

    def to_dot(self):
        dot_vertices = []
        dot_edges = []

        for n, v in self.vertices.items():
            dot_vertices.append(
                '  {} [label="{}|{}"]'.format(n, n, v.payload or '')
            )
            for e in v.outgoing_edges:
                label = e.requirement
                dot_edges.append(
                    '  {} -> {} [label="{}"]'.format(
                        e.origin.name,
                        e.destination.name,
                        label
                    )
                )

        dot_vertices = sorted(set(dot_vertices))
        dot_edges = sorted(set(dot_edges))

        dot_vertices.insert(0, 'digraph G {')
        dot_vertices.append('')
        dot_edges.append('}')

        dot = dot_vertices + dot_edges

        return '\n'.join(dot)

    def __iter__(self):
        return iter(self.vertices.values())



import re
from typing import Union

from poetry.packages import Package
from poetry.semver.comparison import less_than
from poetry.semver.helpers import normalize_version
from poetry.semver.version_parser import VersionParser


class VersionSelector(object):

    def __init__(self, pool, parser=VersionParser()):
        self._pool = pool
        self._parser = parser

    def find_best_candidate(self,
                            package_name: str,
                            target_package_version: Union[str, None] = None
                            ) -> Union[Package, bool]:
        """
        Given a package name and optional version,
        returns the latest Package that matches
        """
        if target_package_version:
            constraint = self._parser.parse_constraints(target_package_version)
        else:
            constraint = None

        candidates = self._pool.find_packages(package_name, constraint)

        if not candidates:
            return False

        # Select highest version if we have many
        package = candidates[0]
        for candidate in candidates:
            # Select highest version of the two
            if less_than(package.version, candidate.version):
                package = candidate

        return package

    def find_recommended_require_version(self, package):
        version = package.version

        return self._transform_version(version, package.pretty_version)

    def _transform_version(self, version, pretty_version):
        # attempt to transform 2.1.1 to 2.1
        # this allows you to upgrade through minor versions
        try:
            parts = normalize_version(version).split('.')
        except ValueError:
            return pretty_version

        # check to see if we have a semver-looking version
        if len(parts) == 4 and re.match('^0\D?', parts[3]):
            # remove the last parts (the patch version number and any extra)
            if parts[0] == '0':
                del parts[3]
            else:
                del parts[3]
                del parts[2]

            version = '.'.join(parts)
        else:
            return pretty_version

        return f'^{version}'



from . import factory, abstracttable
from .common import Element


class InlineTableElement(abstracttable.AbstractTable):
    """
    An Element containing key-value pairs, representing an inline table.

    Implements dict-like interface.

    Assumes input sub_elements are correct for an inline table element.
    """

    def __init__(self, sub_elements):
        abstracttable.AbstractTable.__init__(self, sub_elements)

    def __setitem__(self, key, value):

        new_element = value if isinstance(value, Element) else factory.create_element(value)

        try:

            key_i, value_i = self._find_key_and_value(key)
            # Found, then replace the value element with a new one
            self._sub_elements = self.sub_elements[:value_i] + [new_element] + self.sub_elements[value_i+1:]

        except KeyError:    # Key does not exist, adding anew!

            new_entry = [
                factory.create_string_element(key, bare_allowed=True),
                factory.create_whitespace_element(),
                factory.create_operator_element('='),
                factory.create_whitespace_element(),
                new_element,
            ]

            if self:    # If not empty
                new_entry = [
                    factory.create_operator_element(','),
                    factory.create_whitespace_element(),
                ] + new_entry

            insertion_index = self._find_closing_curly_bracket()
            self._sub_elements = self.sub_elements[:insertion_index] + new_entry + self.sub_elements[insertion_index:]

    def __delitem__(self, key):

        key_i, value_i = self._find_key_and_value(key)

        begin, end = key_i, value_i+1

        # Rules:
        #   1. begin should be index to the preceding comma to the key
        #   2. end should be index to the following comma, or the closing bracket
        #   3. If no preceding comma found but following comma found then end should be the index of the following key

        preceding_comma = self._find_preceding_comma(begin)
        found_preceding_comma = preceding_comma >= 0
        if found_preceding_comma:
            begin = preceding_comma

        following_comma = self._find_following_comma(value_i)
        if following_comma >= 0:
            if not found_preceding_comma:
                end = self._find_following_non_metadata(following_comma)
            else:
                end = following_comma
        else:
            end = self._find_closing_curly_bracket()

        self._sub_elements = self.sub_elements[:begin] + self.sub_elements[end:]

    def multiline_equivalent(self):
        return factory.create_inline_table(self.primitive_value, multiline_table=True, multiline_strings_allowed=True)

    @property
    def value(self):
        return self     # self is a dict-like value that is perfectly usable



import operator

from poetry.version import parse as parse_version
from poetry.version import version_compare

from ..helpers import normalize_version
from .base_constraint import BaseConstraint


class Constraint(BaseConstraint):

    OP_EQ = operator.eq
    OP_LT = operator.lt
    OP_LE = operator.le
    OP_GT = operator.gt
    OP_GE = operator.ge
    OP_NE = operator.ne

    _trans_op_str = {
        '=': OP_EQ,
        '==': OP_EQ,
        '<': OP_LT,
        '<=': OP_LE,
        '>': OP_GT,
        '>=': OP_GE,
        '!=': OP_NE
    }

    _trans_op_int = {
        OP_EQ: '==',
        OP_LT: '<',
        OP_LE: '<=',
        OP_GT: '>',
        OP_GE: '>=',
        OP_NE: '!='
    }

    def __init__(self, operator: str, version: str):
        if operator not in self.supported_operators:
            raise ValueError(
                f'Invalid operator "{operator}" given, '
                f'expected one of: {", ".join(self.supported_operators)}'
            )

        self._operator = self._trans_op_str[operator]
        self._string_operator = operator
        self._version = str(parse_version(version))
        
    @property
    def supported_operators(self) -> list:
        return list(self._trans_op_str.keys())

    @property
    def operator(self):
        return self._operator

    @property
    def string_operator(self):
        return self._string_operator

    @property
    def version(self) -> str:
        return self._version

    def matches(self, provider):
        if (
            isinstance(provider, self.__class__)
            and provider.__class__ is self.__class__
        ):
            return self.match_specific(provider)

        # turn matching around to find a match
        return provider.matches(self)

    def version_compare(self, a: str, b: str, operator: str) -> bool:
        if operator not in self._trans_op_str:
            raise ValueError(
                f'Invalid operator "{operator}" given, '
                f'expected one of: {", ".join(self.supported_operators)}'
            )

        # If we can't normalize the version
        # we delegate to parse_version()
        try:
            a = normalize_version(a)
        except ValueError:
            pass

        try:
            b = normalize_version(b)
        except ValueError:
            pass

        return version_compare(a, b, operator)

    def match_specific(self, provider: 'Constraint') -> bool:
        no_equal_op = self._trans_op_int[self._operator].replace('=', '')
        provider_no_equal_op = self._trans_op_int[provider.operator].replace('=', '')

        is_equal_op = self.OP_EQ is self._operator
        is_non_equal_op = self.OP_NE is self._operator
        is_provider_equal_op = self.OP_EQ is provider.operator
        is_provider_non_equal_op = self.OP_NE is provider.operator

        # '!=' operator is match when other operator
        # is not '==' operator or version is not match
        # these kinds of comparisons always have a solution
        if is_non_equal_op or is_provider_non_equal_op:
            return (not is_equal_op and not is_provider_equal_op
                    or self.version_compare(provider.version,
                                            self._version,
                                            '!='))

        # An example for the condition is <= 2.0 & < 1.0
        # These kinds of comparisons always have a solution
        if (self._operator is not self.OP_EQ
                and no_equal_op == provider_no_equal_op):
            return True

        if self.version_compare(
            provider.version,
            self.version,
            self._trans_op_int[self._operator]
        ):
            # special case, e.g. require >= 1.0 and provide < 1.0
            # 1.0 >= 1.0 but 1.0 is outside of the provided interval
            if (
                provider.version == self.version
                and self._trans_op_int[provider.operator] == provider_no_equal_op
                and self._trans_op_int[self.operator] != no_equal_op
            ):
                return False

            return True

        return False

    def __str__(self):
        return '{} {}'.format(
            self._trans_op_int[self._operator],
            self._version
        )

    def __repr__(self):
        return '<Constraint \'{}\'>'.format(str(self))



import json

import poetry.packages

from hashlib import sha256
from pathlib import Path
from typing import List

from poetry.repositories import Repository
from poetry.utils.toml_file import TomlFile


class Locker:

    _relevant_keys = [
        'name',
        'version',
        'python-versions',
        'platform',
        'dependencies',
        'dev-dependencies',
        'source',
    ]

    def __init__(self, lock: Path, local_config: dict):
        self._lock = TomlFile(lock)
        self._local_config = local_config
        self._lock_data = None
        self._content_hash = self._get_content_hash()

    @property
    def lock(self) -> TomlFile:
        return self._lock

    @property
    def lock_data(self):
        if self._lock_data is None:
            self._lock_data = self._get_lock_data()

        return self._lock_data

    def is_locked(self) -> bool:
        """
        Checks whether the locker has been locked (lockfile found).
        """
        if not self._lock.exists():
            return False

        return 'package' in self.lock_data

    def is_fresh(self) -> bool:
        """
        Checks whether the lock file is still up to date with the current hash.
        """
        lock = self._lock.read(True)
        metadata = lock.get('metadata', {})

        if 'content-hash' in metadata:
            return self._content_hash == lock['metadata']['content-hash']

        return False

    def locked_repository(self, with_dev_reqs: bool = False) -> Repository:
        """
        Searches and returns a repository of locked packages.
        """
        if not self.is_locked():
            return Repository()

        lock_data = self.lock_data
        packages = Repository()

        if with_dev_reqs:
            locked_packages = lock_data['package']
        else:
            locked_packages = [
                p for p in lock_data['package'] if p['category'] == 'main'
            ]

        if not locked_packages:
            return packages

        for info in locked_packages:
            package = poetry.packages.Package(
                info['name'],
                info['version'],
                info['version']
            )
            package.description = info.get('description', '')
            package.category = info['category']
            package.optional = info['optional']
            package.hashes = lock_data['metadata']['hashes'][info['name']]
            package.python_versions = info['python-versions']

            for dep_name, constraint in info.get('dependencies', {}).items():
                package.add_dependency(dep_name, constraint)

            if 'requirements' in info:
                package.requirements = info['requirements']

            if 'source' in info:
                package.source_type = info['source']['type']
                package.source_url = info['source']['url']
                package.source_reference = info['source']['reference']

            packages.add_package(package)

        return packages

    def set_lock_data(self,
                      root, packages) -> bool:
        hashes = {}
        packages = self._lock_packages(packages)
        # Retrieving hashes
        for package in packages:
            hashes[package['name']] = package['hashes']
            del package['hashes']

        lock = {
            'package': packages,
            'metadata': {
                'python-versions': root.python_versions,
                'platform': root.platform,
                'content-hash': self._content_hash,
                'hashes': hashes,
            }
        }

        if root.extras:
            lock['extras'] = {
                extra: [dep.pretty_name for dep in deps]
                for extra, deps in root.extras.items()
            }

        if not self.is_locked() or lock != self.lock_data:
            self._write_lock_data(lock)

            return True

        return False

    def _write_lock_data(self, data):
        self._lock.write(data)
        self._lock_data = None

    def _get_content_hash(self) -> str:
        """
        Returns the sha256 hash of the sorted content of the composer file.
        """
        content = self._local_config

        relevant_content = {}
        for key in self._relevant_keys:
            relevant_content[key] = content.get(key)

        content_hash = sha256(
            json.dumps(relevant_content, sort_keys=True).encode()
        ).hexdigest()

        return content_hash

    def _get_lock_data(self) -> dict:
        if not self._lock.exists():
            raise RuntimeError(
                'No lockfile found. Unable to read locked packages'
            )

        return self._lock.read(True)

    def _lock_packages(self,
                       packages: List['poetry.packages.Package']) -> list:
        locked = []

        for package in sorted(packages, key=lambda x: x.name):
            spec = self._dump_package(package)

            locked.append(spec)

        return locked

    def _dump_package(self, package: 'poetry.packages.Package') -> dict:
        dependencies = {}
        for dependency in package.requires:
            if dependency.is_optional():
                continue

            dependencies[dependency.pretty_name] = str(dependency.pretty_constraint)

        data = {
            'name': package.pretty_name,
            'version': package.pretty_version,
            'description': package.description,
            'category': package.category,
            'optional': package.optional,
            'python-versions': package.python_versions,
            'platform': package.platform,
            'hashes': package.hashes,
            'dependencies': dependencies
        }

        if package.source_type:
            data['source'] = {
                'type': package.source_type,
                'url': package.source_url,
                'reference': package.source_reference
            }

        if package.requirements:
            data['requirements'] = package.requirements

        return data




from cleo.styles import CleoStyle
from cleo.styles import OutputStyle


class PoetryStyle(CleoStyle):

    def __init__(self, i, o):
        super().__init__(i, o)

        self.output.get_formatter().add_style('warning', 'black', 'yellow')
        self.output.get_formatter().add_style('question', 'blue')

    def writeln(self, messages,
                type=OutputStyle.OUTPUT_NORMAL,
                verbosity=OutputStyle.VERBOSITY_NORMAL):
        if self.output.verbosity >= verbosity:
            super().writeln(messages, type=type)

    def write(self, messages,
              newline=False,
              type=OutputStyle.OUTPUT_NORMAL,
              verbosity=OutputStyle.VERBOSITY_NORMAL):
        if self.output.verbosity >= verbosity:
            super().write(messages, newline=newline, type=type)



from .add_edge_no_circular import AddEdgeNoCircular
from .add_vertex import AddVertex
from .delete_edge import DeleteEdge
from .detach_vertex_named import DetachVertexNamed
from .set_payload import SetPayload
from .tag import Tag


class Log:
    """
    A log for dependency graph actions.
    """

    def __init__(self):
        self._current_action = None
        self._first_action = None

    def tag(self, graph, tag):
        """
        Tags the current state of the dependency as the given tag.
        """
        return self._push_action(graph, Tag(tag))

    def add_vertex(self, graph, name, payload, root):
        return self._push_action(graph, AddVertex(name, payload, root))

    def detach_vertex_named(self, graph, name):
        return self._push_action(graph, DetachVertexNamed(name))

    def add_edge_no_circular(self, graph, origin, destination, requirement):
        action = AddEdgeNoCircular(origin, destination, requirement)
        return self._push_action(graph, action)

    def delete_edge(self, graph, origin, destination, requirement):
        action = DeleteEdge(origin, destination, requirement)
        return self._push_action(graph, action)

    def set_payload(self, graph, name, payload):
        return self._push_action(graph, SetPayload(name, payload))

    def pop(self, graph):
        action = self._current_action
        if not action:
            return

        self._current_action = action.previous
        if not self._current_action:
            self._first_action = None

        action.down(graph)

        return action

    def rewind_to(self, graph, tag):
        while True:
            action = self.pop(graph)
            if not action:
                raise ValueError('No tag "{}" found'.format(tag))

            if isinstance(action, Tag) and action.tag == tag:
                break

    def _push_action(self, graph, action):
        """
        Adds the given action to the log, running the action

        :param graph: The graph
        :param action: The action
        :type action: Action
        """
        action.previous = self._current_action
        if self._current_action:
            self._current_action.next = action

        self._current_action = action
        if not self._first_action:
            self._first_action = action

        return action.up(graph)



from typing import Any


class Action:

    def __init__(self):
        self.previous = None
        self.next = None

    @property
    def action_name(self) -> str:
        raise NotImplementedError()

    def up(self, graph: 'DependencyGraph') -> Any:
        """
        Performs the action on the given graph.
        """
        raise NotImplementedError()

    def down(self, graph: 'DependencyGraph') -> None:
        """
        Reverses the action on the given graph.
        """
        raise NotImplementedError()



import toml

from pathlib import Path

from poetry.toml import dumps
from poetry.toml import loads
from poetry.toml import TOMLFile


class TomlFile:

    def __init__(self, path):
        self._path = Path(path)

    @property
    def path(self):
        return self._path

    def read(self, raw=False) -> dict:
        if raw:
            return toml.loads(self._path.read_text())

        return loads(self._path.read_text())

    def write(self, data) -> None:
        if not isinstance(data, TOMLFile):
            data = toml.dumps(data)
        else:
            data = dumps(data)

        self._path.write_text(data)

    def __getattr__(self, item):
        return getattr(self._path, item)



from .action import Action


class DetachVertexNamed(Action):

    def __init__(self, name):
        super(DetachVertexNamed, self).__init__()

        self._name = name
        self._vertex = None

    @property
    def action_name(self):
        return 'detach_vertex'

    @property
    def name(self):
        return self._name

    def up(self, graph):
        if self._name not in graph.vertices:
            return []

        self._vertex = graph.vertices[self._name]
        del graph.vertices[self._name]
        removed_vertices = [self._vertex]
        for e in self._vertex.outgoing_edges:
            v = e.destination
            try:
                v.incoming_edges.remove(e)
            except ValueError:
                pass

            if not v.root and not v.incoming_edges:
                removed_vertices += graph.detach_vertex_named(v.name)

        for e in self._vertex.incoming_edges:
            v = e.origin

            try:
                v.outgoing_edges.remove(e)
            except ValueError:
                pass

        return removed_vertices

    def down(self, graph):
        if self._vertex is None:
            return

        graph.vertices[self._vertex.name] = self._vertex
        for e in self._vertex.outgoing_edges:
            e.destination.incoming_edges.append(e)

        for e in self._vertex.incoming_edges:
            e.origin.outgoing_edges.append(e)



import itertools


def is_sequence_like(x):
    """
    Returns True if x exposes a sequence-like interface.
    """
    required_attrs = (
        '__len__',
        '__getitem__'
    )
    return all(hasattr(x, attr) for attr in required_attrs)


def is_dict_like(x):
    """
    Returns True if x exposes a dict-like interface.
    """
    required_attrs = (
        '__len__',
        '__getitem__',
        'keys',
        'values',
    )
    return all(hasattr(x, attr) for attr in required_attrs)


def join_with(iterable, separator):
    """
    Joins elements from iterable with separator and returns the produced sequence as a list.

    separator must be addable to a list.
    """
    inputs = list(iterable)
    b = []
    for i, element in enumerate(inputs):
        if isinstance(element, (list, tuple, set)):
            b += tuple(element)
        else:
            b += [element]
        if i < len(inputs)-1:
            b += separator
    return b


def chunkate_string(text, length):
    """
    Iterates over the given seq in chunks of at maximally the given length. Will never break a whole word.
    """
    iterator_index = 0

    def next_newline():
        try:
            return next(i for (i, c) in enumerate(text) if i > iterator_index and c == '\n')
        except StopIteration:
            return len(text)

    def next_breaker():
        try:
            return next(i for (i, c) in reversed(tuple(enumerate(text)))
                        if i >= iterator_index and
                        (i < iterator_index+length) and
                        c in (' ', '\t'))
        except StopIteration:
            return len(text)

    while iterator_index < len(text):
        next_chunk = text[iterator_index:min(next_newline(), next_breaker()+1)]
        iterator_index += len(next_chunk)
        yield next_chunk


def flatten_nested(nested_dicts):
    """
    Flattens dicts and sequences into one dict with tuples of keys representing the nested keys.

    Example
    >>> dd = { \
        'dict1': {'name': 'Jon', 'id': 42}, \
        'dict2': {'name': 'Sam', 'id': 41}, \
        'seq1': [{'one': 1, 'two': 2}] \
        }

    >>> flatten_nested(dd) == { \
        ('dict1', 'name'): 'Jon', ('dict1', 'id'): 42, \
        ('dict2', 'name'): 'Sam', ('dict2', 'id'): 41, \
        ('seq1', 0, 'one'): 1, ('seq1', 0, 'two'): 2, \
        }
    True
    """
    assert isinstance(nested_dicts, (dict, list, tuple)), 'Only works with a collection parameter'

    def items(c):
        if isinstance(c, dict):
            return c.items()
        elif isinstance(c, (list, tuple)):
            return enumerate(c)
        else:
            raise RuntimeError('c must be a collection')

    def flatten(dd):
        output = {}
        for k, v in items(dd):
            if isinstance(v, (dict, list, tuple)):
                for child_key, child_value in flatten(v).items():
                    output[(k,) + child_key] = child_value
            else:
                output[(k,)] = v
        return output

    return flatten(nested_dicts)


class PeekableIterator:

    # Returned by peek() when the iterator is exhausted. Truthiness is False.
    Nothing = tuple()

    def __init__(self, iter):
        self._iter = iter

    def __next__(self):
        return next(self._iter)

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self

    def peek(self):
        """
        Returns PeekableIterator.Nothing when the iterator is exhausted.
        """
        try:
            v = next(self._iter)
            self._iter = itertools.chain((v,), self._iter)
            return v
        except StopIteration:
            return PeekableIterator.Nothing
